import pandas as pd
import numpy as np
from .core import rga
from .util.utils import manipulate_testdata, validate_variables, convert_to_dataframe



def compute_single_variable_rge(xtrain, xtest, yhat, model, variables):
    """Compute RGE for single variable contribution."""
    xtrain, xtest, yhat = convert_to_dataframe(xtrain, xtest, yhat)
    validate_variables(variables, xtrain)
    rge_list = []
    for i in variables:
        xtest_rm = manipulate_testdata(xtrain, xtest, model, i)
        yhat_rm = [x[1] for x in model.predict_proba(xtest_rm)]
        rge_list.append(1 - (rga(yhat, yhat_rm)))
    rge_df = pd.DataFrame(rge_list, index=variables, columns=["RGE"]).sort_values(by="RGE", ascending=False)
    return rge_df


def compute_group_variable_rge(xtrain, xtest, yhat, model, variables):
    """Compute RGE for group variable contribution."""
    xtrain, xtest, yhat = convert_to_dataframe(xtrain, xtest, yhat)
    validate_variables(variables, xtrain)
    for i in variables:
        xtest_rm = manipulate_testdata(xtrain, xtest, model, i)
    yhat_rm = [x[1] for x in model.predict_proba(xtest_rm)]
    rge = 1 - (rga(yhat, yhat_rm))
    rge_df = pd.DataFrame(rge, index=[str(variables)], columns=["RGE"])
    return rge_df


def compute_full_single_rge(xtrain, xtest, yhat, model):
    """Compute RGE when no variables are specified."""
    xtrain, xtest, yhat = convert_to_dataframe(xtrain, xtest, yhat)
    rge_list = []
    for i in xtest.columns:
        xtest_rm = manipulate_testdata(xtrain, xtest, model, i)
        yhat_rm = [x[1] for x in model.predict_proba(xtest_rm)]
        rge_list.append(1 - (rga(yhat, yhat_rm)))
    rge_df = pd.DataFrame(rge_list, index= xtest.columns, columns=["RGE"]).sort_values(by="RGE", ascending=False)
    return rge_df





import pandas as pd
import numpy as np

def manipulate_testdata(xtrain, xtest, model, variable):
    # create xtest_rm
    xtest_rm = xtest.copy()
    cat_indices = model.get_cat_feature_indices()
    # replace variable with mode or mean based on its type
    if xtrain.columns.get_loc(variable) not in cat_indices:
        mean_value = xtrain[variable].mean()
        xtest_rm[variable] = mean_value
    else:
        mode_value = xtrain[variable].mode()[0]
        xtest_rm[variable] = mode_value
    return xtest_rm

def convert_to_dataframe(*args):
    """Convert inputs to DataFrames."""
    return [pd.DataFrame(arg).reset_index(drop=True) for arg in args]

def validate_variables(variables, xtrain):
    """Check if variables are valid and exist in the train dataset."""
    if not isinstance(variables, list):
        raise ValueError("Variables input must be a list")
    for var in variables:
        if var not in xtrain.columns:
            raise ValueError(f"{var} is not in the variables")



import pandas as pd
import numpy as np


def rga(y, yhat):
        """
        RANK GRADUATION ACCURACY (RGA) MEASURE 
        Function for the RGA measure computation

        Inputs:
        y    : Actual values of the target variable 
        yhat : Predicted values generated by the model 
        
        Returns:
        RGA  : Calculated RGA measure
        """
        # Convert inputs to DataFrames and concatenate them
        y = pd.DataFrame(y).reset_index(drop=True)
        yhat = pd.DataFrame(yhat).reset_index(drop=True)

        if y.shape != yhat.shape:
                raise ValueError("y and yhat should have the same shape.")

        df = pd.concat([y, yhat], axis=1)
        df.columns = ["y", "yhat"]

        if df.isna().sum().sum() > 0:
                raise ValueError("There are missing values.")
              
        # Rank yhat values
        df['ryhat'] = df['yhat'].rank(method="min")

        # Group by ryhat and calculate mean of y (support)
        support = df.groupby('ryhat')['y'].mean().reset_index(name='support')

        # Merge support back to the original dataframe
        df = pd.merge(df, support, on="ryhat", how="left")

        # Create the rord column by directly assigning 'support' where ryhat matches
        df['rord'] = df['support']
        
        # Sort df by yhat to get correct ordering for ystar
        df = df.sort_values(by="yhat").reset_index(drop=True)

        # Get ystar in the same order as rord in the sorted dataframe
        ystar = df['rord'].values

        # Create the index array I
        I = np.arange(len(df))

        # Calculate conc, dec (descending order of y) and inc (ascending order of y)
        conc = np.sum(I * ystar)
        sorted_y = np.sort(df['y'])  # y sorted in ascending order
        dec = np.sum(I * sorted_y[::-1])  # y sorted in descending order
        inc = np.sum(I * sorted_y)

        # Compute the RGA
        RGA = (conc - dec) / (inc - dec)

        return RGA


import pandas as pd
import numpy as np
from .core import rga
from .util.utils import convert_to_dataframe


def compute_rga_parity(xtrain, xtest, ytest, yhat, model, protectedvariable):

    if protectedvariable not in xtrain.columns:
        raise ValueError(f"{protectedvariable} is not in the variables")
    xtrain, xtest, ytest, yhat = convert_to_dataframe(xtrain, xtest, ytest, yhat)
    protected_groups = xtrain[protectedvariable].value_counts().index
    rgas = []
    for i in protected_groups:
        xtest_pr = xtest[xtest[protectedvariable]== i]
        ytest_pr = ytest.loc[xtest_pr.index]
        yhat_pr = [x[1] for x in model.predict_proba(xtest_pr)]
        rga_value = rga(ytest_pr, yhat_pr)
        rgas.append(rga_value)            
    return f"The RGA-based imparity between the protected gorups is {max(rgas)-min(rgas)}."    

